//
//  NSView+Exts.swift
//  Queen
//
//  Created by 聂子 on 2019/7/6.
//  Copyright © 2019 聂子. All rights reserved.
//

import Cocoa

extension TypeWrapperProtocol where WrappedType: NSView {

    public func removeAllSubviews() {
        self.wrappedValue.subviews.forEach { (v) in
            v.removeFromSuperview()
        }
    }

    public func animationUpdate() {
        self.wrappedValue.updateConstraintsForSubtreeIfNeeded()
        self.wrappedValue.updateConstraints()
        self.wrappedValue.layoutSubtreeIfNeeded()
    }

    /// mouse in view
    ///
    /// - Returns: true / false
    public func mouseInView() -> Bool {
        if self.wrappedValue.window == nil {
            return false
        }
        if self.wrappedValue.isHidden {
            return false
        }
        var point = NSEvent.mouseLocation
        point = self.wrappedValue.window?.convertFromScreen(NSRect.init(origin: point, size: CGSize.init(width: 0, height: 0))).origin ?? CGPoint.init(x: 0, y: 0)
        point = self.wrappedValue.convert(point, from: nil)
        return NSPointInRect(point, self.wrappedValue.visibleRect)
    }

    
    /// 判断 view 是否相交
    /// - Parameter view: 目标view
    /// - Returns: 结果
    public func intersect(with view: NSView) -> Bool {
        let window = self.wrappedValue.window
        
        let selfRect = self.wrappedValue.convert(self.wrappedValue.bounds, to: window?.contentView)
        let viewRect = view.convert(view.bounds, to: window?.contentView)
        return selfRect.intersects(viewRect)
    }
    
    
}

extension NSView {
    // TODO: 由于部分sub class 有自己的backgroundColor 属性，不然会冲突掉
    public var background_color: NSColor? {
        get {
            if let colorRef = self.layer?.backgroundColor {
                return NSColor(cgColor: colorRef)
            } else {
                return nil
            }
        }
        set {
            self.wantsLayer = true
            self.layer?.backgroundColor = newValue?.cgColor
        }
    }
}



extension NSView {
    // x
    public var x : CGFloat {

        get {

            return frame.origin.x
        }

        set(newVal) {

            var tmpFrame : CGRect = frame
            tmpFrame.origin.x     = newVal
            frame                 = tmpFrame
        }
    }

    // y
    public var y : CGFloat {

        get {

            return frame.origin.y
        }

        set(newVal) {

            var tmpFrame : CGRect = frame
            tmpFrame.origin.y     = newVal
            frame                 = tmpFrame
        }
    }

    // height
    public var height : CGFloat {

        get {

            return frame.size.height
        }

        set(newVal) {

            var tmpFrame : CGRect = frame
            tmpFrame.size.height  = newVal
            frame                 = tmpFrame
        }
    }

    // width
    public var width : CGFloat {

        get {

            return frame.size.width
        }

        set(newVal) {

            var tmpFrame : CGRect = frame
            tmpFrame.size.width   = newVal
            frame                 = tmpFrame
        }
    }

    // left
    public var left : CGFloat {

        get {

            return x
        }

        set(newVal) {

            x = newVal
        }
    }

    // right
    public var right : CGFloat {

        get {

            return x + width
        }

        set(newVal) {

            x = newVal - width
        }
    }

    // top
    public var top : CGFloat {

        get {

            return y
        }

        set(newVal) {

            y = newVal
        }
    }

    // bottom
    public var bottom : CGFloat {

        get {

            return y + height
        }

        set(newVal) {

            y = newVal - height
        }
    }

    public var middleX : CGFloat {

        get {

            return width / 2
        }
    }

    public var middleY : CGFloat {

        get {

            return height / 2
        }
    }

    public var middlePoint : CGPoint {

        get {

            return CGPoint(x: middleX, y: middleY)
        }
    }

    public var maxX: CGFloat {
        get {
            return frame.maxX
        }
    }

    public var maxY: CGFloat {
        get {
            return frame.maxY
        }
    }

}




extension NSView {

    public var origin: CGPoint {
        get {
            frame.origin
        }
        set {
            frame.origin.x = newValue.x
            frame.origin.y = newValue.y
        }
    }

    public var size: CGSize {
        get {
            frame.size
        }
        set {
            width = newValue.width
            height = newValue.height
        }
    }


    @IBInspectable
    public var borderColor: NSColor? {
        get {
            guard let color = layer?.borderColor else { return nil }
            return NSColor(cgColor: color)
        }
        set {
            wantsLayer = true
            layer?.borderColor = newValue?.cgColor
        }
    }

    @IBInspectable
    public var borderWidth: CGFloat {
        get {
            layer?.borderWidth ?? 0
        }
        set {
            wantsLayer = true
            layer?.borderWidth = newValue
        }
    }

    @IBInspectable
    public var cornerRadius: CGFloat {
        get {
            layer?.cornerRadius ?? 0
        }
        set {
            wantsLayer = true
            layer?.masksToBounds = true
            layer?.cornerRadius = abs(CGFloat(Int(newValue * 100)) / 100)
        }
    }

    @IBInspectable
    public var shadowColor: NSColor? {
        get {
            guard let color = layer?.shadowColor else { return nil }
            return NSColor(cgColor: color)
        }
        set {
            wantsLayer = true
            layer?.shadowColor = newValue?.cgColor
        }
    }

    @IBInspectable
    public var shadowOffset: CGSize {
        get {
            layer?.shadowOffset ?? CGSize.zero
        }
        set {
            wantsLayer = true
            layer?.shadowOffset = newValue
        }
    }

    @IBInspectable
    public var shadowOpacity: Float {
        get {
            layer?.shadowOpacity ?? 0
        }
        set {
            wantsLayer = true
            layer?.shadowOpacity = newValue
        }
    }

    @IBInspectable
    public var shadowRadius: CGFloat {
        get {
            return layer?.shadowRadius ?? 0
        }
        set {
            wantsLayer = true
            layer?.shadowRadius = newValue
        }
    }
}






public final class ObjectAssociation<T: AnyObject> {

    private let policy: objc_AssociationPolicy

    /// - Parameter policy: An association policy that will be used when linking objects.
    public init(policy: objc_AssociationPolicy = .OBJC_ASSOCIATION_RETAIN_NONATOMIC) {

        self.policy = policy
    }

    /// Accesses associated object.
    /// - Parameter index: An object whose associated object is to be accessed.
    public subscript(index: AnyObject) -> T? {

        get { return objc_getAssociatedObject(index, Unmanaged.passUnretained(self).toOpaque()) as! T? }
        set { objc_setAssociatedObject(index, Unmanaged.passUnretained(self).toOpaque(), newValue, policy) }
    }
}
