//
//  TKButton.swift
//  TKCocoaModule
//
//  Created by 🐶 on 12/22/21.
//

import AppKit
import Cocoa

public class TKButton: TKControl {
    private let backgroundImageLayer = CALayer.init()
    private let imageLayer = CALayer.init()
    private let textLayer = VerticallyCenteredTextLayer.init()
    
    
    
    private var textsDic: [TKControl.State: NSAttributedString] = [:]
    private var imageDic:[TKControl.State: NSImage] = [:]
    private var backImageDic:[TKControl.State: NSImage] = [:]
    private var textColorsDic:[TKControl.State: NSColor] = [:]
    private var fontsDic:[TKControl.State: NSFont] = [:]
    private var titleShadowColorDics:[TKControl.State: NSColor] = [:]
    
    
    public override var wantsLayer: Bool {
        set {
            super.wantsLayer = true
        }
        get {
            return true
        }
    }
    
    public override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        
        self.wantsLayer = true
        
        self.backgroundImageLayer.name = LayerName.backgroundImageLayer
        self.imageLayer.name = LayerName.imageLayer
        self.textLayer.name = LayerName.textLayer
        
        self.layer?.addSublayer(self.backgroundImageLayer)
        self.layer?.addSublayer(self.textLayer)
        self.layer?.addSublayer(self.imageLayer)
        
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    private var needUpdateLayerConstraints: Bool = true
    public var imagePosition: NSControl.ImagePosition = .imageLeft {
        didSet {
            needUpdateLayerConstraints = true
            self.needsDisplay = true
        }
    }
    
    public var textAlignment: NSTextAlignment = .center {
        didSet {
            needUpdateLayerConstraints = true
            self.needsDisplay = true
        }
    }
    
    public var contentEdget:NSEdgeInsets = .zero {
        didSet {
            needUpdateLayerConstraints = true
            self.needsDisplay = true
        }
    }
    public var titleEdgeInsets: NSEdgeInsets =  .zero {
        didSet {
            needUpdateLayerConstraints = true
            self.needsDisplay = true
        }
    }
    public var imageEdgeInsets: NSEdgeInsets = .zero {
        didSet {
            needUpdateLayerConstraints = true
            self.needsDisplay = true
        }
    }
}

extension TKButton {
    public override func makeBackingLayer() -> CALayer {
        let layer = CALayer.init()
        layer.layoutManager = CAConstraintLayoutManager.init()
        layer.needsDisplayOnBoundsChange = true
        return layer
    }
    public override var needsDisplay: Bool {
        didSet {
            self.layer?.setNeedsLayout()
            self.layer?.setNeedsDisplay()
        }
    }
}

extension TKButton {
    public var currentTitle: String? {
        return textsDic[state]?.string
    }
    public var currentAttributedTitle: NSAttributedString? {
        return textsDic[state]
    }
    public var currentTitleColor: NSColor {
        return textColorsDic[state] ?? NSColor.black
    }
    public var currentImage: NSImage? {
        return imageDic[state]
    }
    public var currentBackgroundImage: NSImage? {
        return backImageDic[state]
    }
    public var currentFont: NSFont {
        return fontsDic[state] ?? NSFont.systemFont(ofSize: 12)
    }
    public var currentTitleShadowColor: NSColor? {
        return titleShadowColorDics[state]
    }
}


extension TKButton {

    public func setTitle(_ title: String, for state: TKControl.State) {
        textsDic.updateValue(NSMutableAttributedString.init(string: title), forKey: state)
    }
    public func title(for state: TKControl.State) -> String? {
        guard let attributeString = textsDic[state] else {
            return nil
        }
        return attributeString.string
    }

    public func setAttributedTitle(_ title: NSAttributedString, for state: TKControl.State) {
        textsDic.updateValue(NSMutableAttributedString.init(attributedString: title), forKey: state)
    }
    public func attributedTitle(for state: TKControl.State) -> NSAttributedString? {
        guard let attributeString = textsDic[state] else {
            return nil
        }
        return attributeString
    }
    public func setTitleColor(_ color: NSColor?, for state: TKControl.State) {
        textColorsDic.updateValue(color ?? NSColor.black, forKey: state)
    }
    public func titleColor(for state: TKControl.State) -> NSColor? {
        return textColorsDic[state]
    }
    
    public func setTitleShadowColor(_ color: NSColor?, for state: TKControl.State) {
        titleShadowColorDics.updateValue(color ?? NSColor.black, forKey: state)
    }
    public func titleShadowColor(for state: TKControl.State) -> NSColor? {
        return titleShadowColorDics[state]
    }
    public func setTitleFont(_ font: NSFont?, for state: TKControl.State) {
        fontsDic.updateValue(font ?? NSFont.systemFont(ofSize: 12), forKey: state)
    }
    public func titleFont(for state: TKControl.State) -> NSFont? {
        return fontsDic[state]
    }
}


// MARK: image
extension TKButton {
    public func backgroundImage(for state: TKControl.State) -> NSImage? {
        return backImageDic[state]
    }
    public func image(for state: TKControl.State) -> NSImage? {
        return imageDic[state]
    }
    public func setBackgroundImage(_ image: NSImage?, for state: TKControl.State){
        guard let image = image else {
            self.backImageDic.removeValue(forKey: state)
            return
        }
        backImageDic.updateValue(image, forKey: state)
    }
    public func setImage(_ image: NSImage?, for state: TKControl.State){
        guard let image = image else {
            self.imageDic.removeValue(forKey: state)
            return
        }
        imageDic.updateValue(image, forKey: state)
    }
}




extension TKButton {
    
    private struct LayerName {
        static let superLayer = "superlayer"
        static let backgroundImageLayer = "backgroundImageLayer"
        static let imageLayer = "imageLayer"
        static let textLayer = "textLayer"
    }
    
    public override var wantsUpdateLayer: Bool {
        return true
    }
    
    public override func layout() {
        super.layout()
        self.layer?.addSublayer(self.backgroundImageLayer)
        self.layer?.addSublayer(self.textLayer)
        self.layer?.addSublayer(self.imageLayer)
        if backImageDic.isEmpty {
            self.backgroundImageLayer.removeFromSuperlayer()
        }
        
        
        self.backgroundImageLayer.contents = currentBackgroundImage
        self.backgroundImageLayer.contentsGravity = .center
        
        self.imageLayer.contents = currentImage
        self.imageLayer.contentsGravity = .resizeAspect
        
        self.textLayer.string = currentAttributedTitle
        self.textLayer.font = currentFont
        self.textLayer.fontSize = currentFont.pointSize
        self.textLayer.contentsScale = NSScreen.main?.scale ?? 1
        self.textLayer.foregroundColor = currentTitleColor.cgColor
        self.textLayer.alignmentMode = self.textAlignment.alignmentMode
        self.textLayer.isWrapped = true
        self.textLayer.contentsGravity = .center
        
        
        switch imagePosition {
                
            case .noImage:
                self.imageLayer.removeFromSuperlayer()
            case .imageOnly:
                self.textLayer.removeFromSuperlayer()
                self.imageLayer.contentsGravity = .center
            case .imageLeft:
                break
            case .imageRight, .imageTrailing:
                break
            case .imageBelow,.imageOverlaps:
                self.layer?.insertSublayer(self.imageLayer, below: self.textLayer)
            case .imageAbove:
                self.layer?.insertSublayer(self.textLayer, below: self.imageLayer)
            case .imageLeading:
                break
            @unknown default:
                break
        }
        
        if self.layer?.sublayers?.contains(self.backgroundImageLayer) ?? false {
            self.layer?.cocoa.sendSublayerToBack(self.backgroundImageLayer)
        }
    }
    
    public override func updateLayer() {
        if needUpdateLayerConstraints {
            updateLayerConstrints()
            needUpdateLayerConstraints = false
        }
        super.updateLayer()
    }
    
    private func updateLayerConstrints() {
        guard let subLayers = self.layer?.sublayers else { return }
        
        var imageConstraints:[CAConstraint] = []
        var titleConstraints:[CAConstraint] = []
        

        
        switch imagePosition {
        case .noImage:
            titleConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX, offset: contentEdget.left + titleEdgeInsets.left))
            titleConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY, offset: contentEdget.top + titleEdgeInsets.top))
            titleConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width, offset:-(contentEdget.left + contentEdget.right + titleEdgeInsets.left + titleEdgeInsets.right)))
            titleConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height, offset: -(contentEdget.top + contentEdget.bottom + titleEdgeInsets.top + titleEdgeInsets.bottom)))
            break
        case .imageOnly:
            imageConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX, offset: contentEdget.left + imageEdgeInsets.left))
            imageConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY, offset: contentEdget.top + imageEdgeInsets.top))
            imageConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width, offset: -(contentEdget.left + contentEdget.right + imageEdgeInsets.left + imageEdgeInsets.right)))
            imageConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,  offset: -(contentEdget.top + contentEdget.bottom + imageEdgeInsets.top + imageEdgeInsets.bottom)))
            
            break
        case .imageRight, .imageTrailing:
            var imageWidth:CGFloat = 0
            if let image = currentImage {
                imageWidth = image.size.width
            }
            imageConstraints.append(.init(attribute: .maxX, relativeTo: LayerName.superLayer, attribute: .maxX, offset: -(contentEdget.right + imageEdgeInsets.right)))
            imageConstraints.append(.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY, offset: contentEdget.top + imageEdgeInsets.top))
            
            imageConstraints.append(.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .maxX, offset: -(imageWidth + contentEdget.right + imageEdgeInsets.right)))
            
            imageConstraints.append(.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,  offset: -(contentEdget.top + contentEdget.bottom +  imageEdgeInsets.top + imageEdgeInsets.bottom)))
                
            titleConstraints.append(.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY, offset: contentEdget.top + titleEdgeInsets.top))
            titleConstraints.append(.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,  offset: -(contentEdget.top + contentEdget.bottom + titleEdgeInsets.top + titleEdgeInsets.bottom)))
            titleConstraints.append(.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX, offset: contentEdget.left + titleEdgeInsets.left))
            
            titleConstraints.append(.init(attribute: .maxX, relativeTo: LayerName.imageLayer, attribute: .minX, offset: -(imageEdgeInsets.left + titleEdgeInsets.right)))
            break
            case .imageBelow,.imageOverlaps:
                
                titleConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX,offset: contentEdget.left + titleEdgeInsets.left))
                titleConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + titleEdgeInsets.top))
                titleConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width,offset: -(contentEdget.left + contentEdget.right + titleEdgeInsets.left + titleEdgeInsets.right)))
                titleConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + titleEdgeInsets.top + titleEdgeInsets.bottom)))
                
                imageConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX,offset: contentEdget.left + imageEdgeInsets.left))
                imageConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + imageEdgeInsets.top))
                imageConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width,offset: -(contentEdget.left + contentEdget.right + imageEdgeInsets.left + imageEdgeInsets.right)))
                imageConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + imageEdgeInsets.top + imageEdgeInsets.bottom)))
                
                
            break
        case .imageAbove:
                titleConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX,offset: contentEdget.left + titleEdgeInsets.left))
                titleConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + titleEdgeInsets.top))
                titleConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width,offset: -(contentEdget.left + contentEdget.right + titleEdgeInsets.left + titleEdgeInsets.right)))
                titleConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + titleEdgeInsets.top + titleEdgeInsets.bottom)))
                
                imageConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX,offset: contentEdget.left + imageEdgeInsets.left))
                imageConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + imageEdgeInsets.top))
                imageConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width,offset: -(contentEdget.left + contentEdget.right + imageEdgeInsets.left + imageEdgeInsets.right)))
                imageConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + imageEdgeInsets.top + imageEdgeInsets.bottom)))
                
                
            break
        case .imageLeft, .imageLeading:
                fallthrough
        @unknown default:
            var imageWidth:CGFloat = 0
            if let image = currentImage {
                imageWidth = image.size.width
            }
            imageConstraints.append(CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX,offset: contentEdget.left + imageEdgeInsets.left))
            imageConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + imageEdgeInsets.top))
            imageConstraints.append(CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .minX, offset: imageWidth))
            imageConstraints.append(CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + imageEdgeInsets.top + imageEdgeInsets.bottom)))
            
            titleConstraints.append(CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY,offset: contentEdget.top + titleEdgeInsets.top))
            titleConstraints.append(.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height,offset: -(contentEdget.top + contentEdget.bottom + titleEdgeInsets.top + titleEdgeInsets.bottom)))
            titleConstraints.append(.init(attribute: .minX, relativeTo: LayerName.imageLayer, attribute: .maxX,offset: imageEdgeInsets.right + titleEdgeInsets.left))
            titleConstraints.append(CAConstraint.init(attribute: .maxX, relativeTo: LayerName.superLayer, attribute: .maxX, offset: -(contentEdget.right + titleEdgeInsets.right)))
                break
        }
        
        
        
        if !imageConstraints.isEmpty, subLayers.contains(self.imageLayer) {
            self.imageLayer.constraints = imageConstraints
        }
        if !titleConstraints.isEmpty, subLayers.contains(self.textLayer) {
            self.textLayer.constraints = titleConstraints
        }
        
        if subLayers.contains(self.backgroundImageLayer) {
            self.backgroundImageLayer.constraints = [
                CAConstraint.init(attribute: .minX, relativeTo: LayerName.superLayer, attribute: .minX),
                CAConstraint.init(attribute: .minY, relativeTo: LayerName.superLayer, attribute: .minY),
                CAConstraint.init(attribute: .width, relativeTo: LayerName.superLayer, attribute: .width),
                CAConstraint.init(attribute: .height, relativeTo: LayerName.superLayer, attribute: .height)
            ]
            
        }
    }
}

