//
//  TKControl.swift
//  TKCocoaModule
//
//  Created by 🐶 on 12/21/21.
//
import Cocoa


extension TKControl  {
    public enum Event {
        case upInside
        case down
        case moved
        case exited
        case dragged
        case entered
        case rightUp
        case rightDown
        case rightDragged
        case otherUp
        case otherDown
        case otherDragged
    }
    
    public struct State: OptionSet, Hashable {
        public let rawValue: UInt8
        public init(rawValue: UInt8) {
            self.rawValue = rawValue
        }
        public static let normal    = State(rawValue: 1 << 0)
        public static let highlighted  = State(rawValue: 1 << 1)
        public static let disabled  = State(rawValue: 1 << 2)
        public static let selected = State(rawValue: 1 << 3)
    }

}

private struct ActionSelector: Hashable {
    static func == (lhs: ActionSelector, rhs: ActionSelector) -> Bool {
        guard let lhs_target = lhs.target, let rhs_target = rhs.target else {
            return false
        }
        let lhs_point = Unmanaged.passUnretained(lhs_target).toOpaque()
        let rhs_point = Unmanaged.passUnretained(rhs_target).toOpaque()
        return lhs_point == rhs_point && lhs.selector == rhs.selector
    }
    
    weak var target: AnyObject?
    var selector: Selector
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(selector)
        if let target = target {
            hasher.combine(ObjectIdentifier(target))
        }
        
    }
}



public class TKControl: NSView {
    
    public private(set) var state: TKControl.State = .normal {
        didSet {
            self.needsDisplay = true
        }
    }
    private var actionSelectors:[TKControl.Event: Set<ActionSelector>] = [:]
    private var actionTrackingArea: NSTrackingArea?
    
    public var isEnabled: Bool  = true {
        didSet {
            self.needsDisplay = true
        }
    }
    public var isSelected: Bool = false {
        didSet {
            self.needsDisplay = true
        }
    }
    public var isHighlighted: Bool = false {
        didSet {
            self.needsDisplay = true
        }
    }
    public var keyEquivalent: String = ""
    public var keyEquivalentModifierMask: NSEvent.ModifierFlags = NSEvent.ModifierFlags.init(rawValue: 0)
    public var sound: NSSound?
    
    private var delay: TimeInterval = 0
    private var interval: TimeInterval = 0

    public override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        self.wantsLayer = true
        self.layerContentsRedrawPolicy = .onSetNeedsDisplay
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    

    public override var frame: NSRect {
        didSet {
            if frame.isEmpty, let area = actionTrackingArea {
                self.removeTrackingArea(area)
                return
            }
            trackingArea()
        }
    }
    public override var wantsLayer: Bool {
        set {
            super.wantsLayer = true
        }
        get {
            return true
        }
    }
    
}

extension TKControl : CALayerDelegate {
    public override func makeBackingLayer() -> CALayer {
        let layer = CALayer.init()
        layer.layoutManager = CAConstraintLayoutManager.init()
        layer.needsDisplayOnBoundsChange = true
        return layer
    }
    public override var needsDisplay: Bool {
        didSet {
            self.layer?.setNeedsLayout()
            self.layer?.setNeedsDisplay()
        }
    }
    public func display(_ layer: CALayer) {
        self.layout()
        self.updateLayer()
    }
}

extension TKControl {
    private func trackingArea() {
        let focusTrackingArea = NSTrackingArea.init(rect: self.bounds, options: [.mouseMoved,
                                                                                 .mouseEnteredAndExited,
                                                                                 .activeWhenFirstResponder,
                                                                                 .enabledDuringMouseDrag,
                                                                                 .inVisibleRect,
                                                                                 .cursorUpdate,
                                                                                 .activeInKeyWindow,
                                                                                 .activeInActiveApp,
                                                                                 .activeAlways
                                                                                ],
                                                    owner: self, userInfo: nil)
        addTrackingArea(focusTrackingArea)
        self.actionTrackingArea = focusTrackingArea
    }
}




extension TKControl {
    
    public var allTargets: Set<AnyHashable> {
        let array = actionSelectors.flatMap({$0.value})
        return Set(array.compactMap({$0.target as? AnyHashable}))
    }
    public func actions(forTarget target: AnyObject?, forControlEvent controlEvent: TKControl.Event) -> [Selector]? {
        guard let target = target, let value = actionSelectors[controlEvent]  else {
            return nil
        }
        return value.filter({$0.target?.isEqual(to: target) ?? false}).map({$0.selector})
    }
    public func removeTarget(_ target: AnyObject?, action: Selector?, for controlEvent: TKControl.Event) {
        guard let values = actionSelectors[controlEvent], let target = target, let action = action else {
            return
        }
        
        let newValues = values.filter({$0.selector != action && !($0.target?.isEqual(target) ?? false)})
        actionSelectors.updateValue(newValues, forKey: controlEvent)
    }
    public func addTarget(_ target: AnyObject?, action: Selector, for controlEvent: TKControl.Event) {
        guard var value = actionSelectors[controlEvent] else {
            self.actionSelectors.updateValue([ActionSelector.init(target: target, selector: action)], forKey: controlEvent)
            return
        }
        
        value.insert(ActionSelector.init(target: target, selector: action))
        self.actionSelectors.updateValue(value, forKey: controlEvent)
    }
}
 

extension TKControl {
    
    private func sendAction(_ action: Selector, to target: Any?) {
        NSApp.sendAction(action, to: target, from: self)
    }
    private func sendActions(for controlEvent: TKControl.Event) {
        if let values = actionSelectors[controlEvent] {
            values.forEach { a in
                NSApp.sendAction(a.selector, to: a.target, from: self)
            }
        }
    }
}




extension TKControl {
    open override func mouseUp(with theEvent: NSEvent) {
        super.mouseUp(with: theEvent)
        guard canAction() else {
            return
        }

        if let actions = actionSelectors[.upInside] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }

        if isSelected {
            state = .selected
            return
        }
        state = .normal
    }
    open override func mouseDown(with event: NSEvent) {
        super.mouseDown(with: event)
        guard canAction() else {
            return
        }
        if canHighLight() {
            state = .highlighted
        }
        
        if let sound = sound, !sound.isPlaying {
            sound.play()
        }
        if let actions = actionSelectors[.down] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    
    open override func mouseMoved(with event: NSEvent) {
        super.mouseMoved(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.moved] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    open override func mouseExited(with event: NSEvent) {
        super.mouseExited(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.exited] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    open override func mouseDragged(with event: NSEvent) {
        super.mouseDragged(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.dragged] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    open override func mouseEntered(with event: NSEvent) {
        super.mouseEntered(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.entered] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    open override func rightMouseUp(with event: NSEvent) {
        super.rightMouseUp(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.rightUp] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    public override func rightMouseDown(with event: NSEvent) {
        super.rightMouseDown(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.rightDown] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    
    open override func rightMouseDragged(with event: NSEvent) {
        super.rightMouseDragged(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.rightDragged] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    open override func otherMouseUp(with event: NSEvent) {
        super.otherMouseUp(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.otherUp] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
        
    }
    open override func otherMouseDown(with event: NSEvent) {
        super.otherMouseDown(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.otherDown] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }

    }
    open override func otherMouseDragged(with event: NSEvent) {
        super.otherMouseDragged(with: event)
        guard canAction() else {
            return
        }
        if let actions = actionSelectors[.otherDragged] {
            actions.forEach { s in
                self.sendAction(s.selector, to: s.target)
            }
        }
    }
    
    
    private func canAction() -> Bool {
        if !isEnabled {
            return false
        }
        return true
    }
    
    private func canHighLight() -> Bool {
        return true
    }
}



extension TKControl {
    public func setPeriodicDelay(_ delay: Float, interval: Float) {
        self.delay = TimeInterval(delay)
        self.interval = TimeInterval(interval)
    }
    public func getPeriodicDelay(_ delay: UnsafeMutablePointer<Float>, interval: UnsafeMutablePointer<Float>) {
        delay.pointee = Float(self.delay)
        interval.pointee = Float(self.interval)
    }
    public func performClick(_ sender: Any?) {
        guard canAction() else {
            return
        }
        self.sendActions(for: .upInside)
    }
}


extension TKControl {
    public override func performKeyEquivalent(with event: NSEvent) -> Bool {
        let window = self.window
        let mw = NSApp.modalWindow

        if canAction() && (mw == nil || (window?.worksWhenModal ?? false) || window == mw) {
            let keyEequivalent = self.keyEquivalent
            if !keyEequivalent.isEmpty && keyEequivalent == event.charactersIgnoringModifiers {
                let mask = self.keyEquivalentModifierMask
                let modifiers = event.modifierFlags
                var relevantModifiersMask:NSEvent.ModifierFlags = self.keyEquivalentModifierMask
                if modifiers.contains(NSEvent.ModifierFlags.function) || CharacterSet.controlCharacters.hasMember(inPlane: keyEequivalent.utf8.first ?? 0) {
                    relevantModifiersMask.insert(NSEvent.ModifierFlags.shift)
                }

                if (modifiers.rawValue & relevantModifiersMask.rawValue) == mask.rawValue & relevantModifiersMask.rawValue  {
                    self.performClick(self)
                    return true
                }
            }
        }
        return false
    }
}




