//
//  TKLabel.swift
//  TKCocoaModule
//
//  Created by 🐶 on 12/21/21.
//

import Foundation
import Cocoa

public typealias NSLabel = TKLabel

public class TKLabel: NSView {
    
    public var text: String = "" {
        didSet {
            self.needsDisplay = true
        }
    }
    public var attributeText: NSAttributedString? {
        didSet {
            if text.isEmpty {
                text = attributeText?.string  ?? ""
            }
        }
    }
    
    public var font: NSFont = NSFont.systemFont(ofSize: NSFont.labelFontSize) {
        didSet {
            self.needsDisplay = true
        }
    }
    
    public var textColor: NSColor = NSColor.black {
        didSet {
            self.needsDisplay = true
        }
    }
    public var textAlignment: NSTextAlignment = .left {
        didSet {
            self.needsDisplay = true
        }
    }
    
    public var lineBreakMode: NSLineBreakMode = .byCharWrapping {
        didSet {
            self.needsDisplay = true
        }
    }
    public var adjustsFontSizeToFitWidth: Bool = false {
        didSet {
            self.needsDisplay = true
        }
    }
    public var numberOfLines: Int = 0 {
        didSet {
            self.needsDisplay = true
        }
    }
    
    public var highlightedTextColor: NSColor? {
        didSet {
            self.needsDisplay = true
        }
    }
    public var isHighlighted: Bool = false {
        didSet {
            self.needsDisplay = true
        }
    }
  
    
    public override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        self.wantsLayer = true
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    public override var wantsLayer: Bool {
        set {
            super.wantsLayer = true
        }
        get {
            return true
        }
    }
}

extension TKLabel : CALayerDelegate {
    public override func makeBackingLayer() -> CALayer {
        let layer = CALayer.init()
        layer.needsDisplayOnBoundsChange = true
        layer.delegate = self
        layer.layoutManager = CAConstraintLayoutManager.init()
        layer.contentsScale = NSScreen.main?.scale ?? 1
        return layer
    }
    public override var needsDisplay: Bool {
        didSet {
            self.layer?.setNeedsLayout()
            self.layer?.setNeedsDisplay()
        }
    }
    public func draw(_ layer: CALayer, in ctx: CGContext) {
        _draw(layer, in: ctx)
    }
}


extension TKLabel {
    private var mutableAttreibuteString: NSMutableAttributedString {
        var attrString = NSMutableAttributedString.init(string: self.text)
        if let att = attributeText {
            attrString = NSMutableAttributedString.init(attributedString: att)
        }

        if !attrString.attributeKeys.contains(NSAttributedString.Key.font.rawValue) {
            attrString.addAttributes([NSAttributedString.Key.font: font], range: NSRange.init(location: 0, length: attrString.length))
        }else {
            if let font = attrString.value(forKey: NSAttributedString.Key.font.rawValue) as? NSFont {
                self.font = font
            }
        }
        
        let style = NSMutableParagraphStyle()
        style.alignment = textAlignment
        style.lineBreakMode = self.lineBreakMode
        style.lineSpacing = 5
        if !attrString.attributeKeys.contains(NSAttributedString.Key.paragraphStyle.rawValue) {
            attrString.addAttributes([NSAttributedString.Key.paragraphStyle: style], range: NSRange.init(location: 0, length: attrString.length))
        }
        return attrString
    }
    public override var fittingSize: NSSize {
        let height = numberOfLines != 0 ?  self.font.lineHeight * CGFloat(numberOfLines)  : CGFloat.infinity
        return mutableAttreibuteString.boundingRect(with: NSSize.init(width: CGFloat.infinity, height: height), options: [.usesLineFragmentOrigin , .usesFontLeading]).size + CGSize.init(width: 4, height: 4)
    }
}


extension TKLabel {
    public override var wantsUpdateLayer: Bool {
        return true
    }

    public override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)
    }
    
    private func _draw(_ layer: CALayer, in context: CGContext)  {
        // 设置文本绘制的矩形
        context.textMatrix = .identity

        // 创建一个绘制文本的区域的路径 不必是矩形
        let path = CGMutablePath()

        // 初始化一个矩形路径。
        path.addRect(self.bounds, transform: .identity)

        // 初始化一个字符串
        let attrString = mutableAttreibuteString
        
        var textColor = textColor
        if isHighlighted, let hColor = highlightedTextColor {
            textColor = hColor
        }
        
        attrString.addAttributes([
                                  NSAttributedString.Key.foregroundColor: textColor
                                 ], range: NSRange.init(location: 0, length: attrString.length))
        
        var font = font
        if adjustsFontSizeToFitWidth {
            font = NSFont.init(name: font.fontName, size: adjustsFontToFit(attrString, rect: layer.bounds)) ?? font
        }
        
        attrString.setAttributes([NSAttributedString.Key.font: font], range: NSRange.init(location: 0, length: attrString.length))

        if numberOfLines == 1 {

            let ctFrameSetter = CTFramesetterCreateWithAttributedString(attrString)
            let ctFrame = CTFramesetterCreateFrame(ctFrameSetter, CFRangeMake(0, attrString.length), path, nil)
            let lines = CTFrameGetLines(ctFrame) as NSArray
            var originsArray = [CGPoint](repeating: CGPoint.zero, count: lines.count)
            CTFrameGetLineOrigins(ctFrame, CFRangeMake(0, 0), &originsArray)
            
            for (i,line) in lines.enumerated() {
                var lineAscent:CGFloat      = 0
                var lineDescent:CGFloat     = 0
                var lineLeading:CGFloat     = 0
                let lineWidth = CTLineGetTypographicBounds(line as! CTLine, &lineAscent, &lineDescent, &lineLeading)
                let lineHeight = lineAscent + lineDescent + lineLeading
                var lineOrigin = originsArray[i]

                switch self.textAlignment {
                    case .right:
                        lineOrigin.x = bounds.width - lineWidth > 0 ? bounds.width - lineWidth : 0
                    case .center:
                        lineOrigin.x = (bounds.width - lineWidth) > 0 ? (bounds.width - lineWidth) / 2 : 0
                    case .natural, .left, .justified:
                        fallthrough
                    @unknown default:
                        lineOrigin.x = 0
                }
                lineOrigin.y = bounds.height - lineHeight > 0 ? (bounds.height - lineHeight) / 2  : 0
                context.textPosition = lineOrigin
                CTLineDraw(line as! CTLine, context)
                
                break
           }

        }else {

            let ctFrameSetter = CTFramesetterCreateWithAttributedString(attrString)
            let ctFrame = CTFramesetterCreateFrame(ctFrameSetter, CFRangeMake(0, attrString.length), path, nil)
            let lines = CTFrameGetLines(ctFrame) as NSArray
            var originsArray = [CGPoint](repeating: CGPoint.zero, count: lines.count)
            CTFrameGetLineOrigins(ctFrame, CFRangeMake(0, 0), &originsArray)
            var frameY:CGFloat              = 0
            for (i,line) in lines.enumerated() {
                var lineAscent:CGFloat      = 0
                var lineDescent:CGFloat     = 0
                var lineLeading:CGFloat     = 0
                let lineWidth = CTLineGetTypographicBounds(line as! CTLine, &lineAscent, &lineDescent, &lineLeading)
                var lineOrigin = originsArray[i]
               
                let contentHeight =   (lineLeading + lineDescent + lineAscent) * CGFloat(originsArray.count)

                var yOffset:CGFloat = 0
                switch self.textAlignment {
                    case .right:
                        lineOrigin.x = bounds.width - lineWidth > 0 ? bounds.width - lineWidth : 0
                    case .center:
                        lineOrigin.x = (bounds.width - lineWidth) > 0 ? (bounds.width - lineWidth) / 2 : 0
                        yOffset = bounds.height - contentHeight > 0 ? (bounds.height - contentHeight ) / 2 : 0
                    case .natural, .left, .justified:
                        fallthrough
                    @unknown default:
                        lineOrigin.x = 0
                }
                
               if i > 0{
                   frameY = frameY - lineAscent - lineDescent
                   lineOrigin.y = frameY
               }
               else{
                   frameY = lineOrigin.y - yOffset
                   lineOrigin.y = lineOrigin.y - yOffset
               }
               context.textPosition = lineOrigin
               CTLineDraw(line as! CTLine, context)
            }


            
            
            
            
        }
    }
}


extension String {
    func size(OfFont font: NSFont) -> CGSize {
        return (self as NSString).size(withAttributes: [NSAttributedString.Key.font: font])
    }
}



extension TKLabel {
    private func adjustsFontToFit(_ attriString: NSAttributedString, rect: CGRect) -> CGFloat {
        var fontSize = self.font.pointSize

        var font: NSFont {
            guard let _font = NSFont.init(name: self.font.fontName, size: fontSize) else {
                return self.font
            }
            return _font
        }
        
        let mutableAttString = NSMutableAttributedString.init(attributedString: attriString)
        var stringSize: CGSize {
            mutableAttString.setAttributes([NSAttributedString.Key.font: font], range: NSRange.init(location: 0, length: mutableAttString.length))
            if numberOfLines != 0 {
                return mutableAttString.boundingRect(with: NSSize.init(width: CGFloat.infinity, height: font.lineHeight * CGFloat(numberOfLines)), options: [.usesLineFragmentOrigin]).size
            }else {
                return mutableAttString.boundingRect(with: NSSize.init(width: rect.width, height: CGFloat.infinity), options: [.usesLineFragmentOrigin]).size
            }
        }
        
        
        
        let inset: CGFloat = 2
        // Decreases the font size until criteria met
        var _stringSize = stringSize
        if numberOfLines != 0 {
            while frame.width < stringSize.width + inset {
                if fontSize < 0 {
                    break
                }
                fontSize -= 1
                _stringSize = stringSize
            }
        }else {
            while frame.width < _stringSize.height + inset {
                if fontSize < 0 {
                    break
                }
                fontSize -= 1
                _stringSize = stringSize
            }
        }

        return fontSize
    }
}

extension NSFont {
    var lineHeight: CGFloat {
        return self.fontDescriptor.object(forKey: NSFontDescriptor.AttributeName.size) as? CGFloat  ?? 0
    }
}
